
Function.prototype.bind = function(scope) {
  var _function = this;
  return function() {
    return _function.apply(scope, arguments);
  };
};


// Cross browser, backward compatible solution
(function( window, Date ) {
// feature testing
var raf = window.mozRequestAnimationFrame    ||
          window.webkitRequestAnimationFrame ||
          window.msRequestAnimationFrame     ||
          window.oRequestAnimationFrame;

window.animLoop = function( render, element ) {
  var running, lastFrame = +new Date;
  function loop( now ) {
    if ( running !== false ) {
      raf ?
        raf( loop, element ) :
        // fallback to setTimeout
        setTimeout( loop, 16 );
      // Make sure to use a valid time, since:
      // - Chrome 10 doesn't return it at all
      // - setTimeout returns the actual timeout
      now = now && now > 1E4 ? now : +new Date;
      var deltaT = now - lastFrame;
      // do not render frame when deltaT is too high
      if ( deltaT < 160 ) {
        running = render( deltaT, lastFrame = now );
      }
    }
  }
  loop();
};
})( window, Date );


var Fortune500 = function() {
	
	var me = this;

	me.minYear = 1955;
	me.maxYear = 2011;
	me.cutTop500 = true;

	me.highlightLosses = $('#losses')[0].checked;

	me.curSelection = 0;
	me.nextSelection = 0;
	me.selectionProgress = 0;

	me.deflators = [26.8, 27.2, 28.1, 28.9, 29.1, 29.6, 29.9, 30.2, 30.6, 31, 31.5, 32.4, 33.4, 34.8, 36.7, 38.8, 40.5, 41.8, 44.4, 49.3, 53.8, 56.9, 60.6, 65.2, 72.6, 82.4, 90.9, 96.5, 99.6, 103.9, 107.6, 109.6, 113.6, 118.3, 124, 130.7, 136.2, 140.3, 144.5, 148.2, 152.4, 156.9, 160.5, 163, 166.6, 172.2, 177.1, 179.9, 184, 188.9, 195.3, 201.6, 207.3, 215.303, 214.537, 218.056];
	
	me.curLUT = [];

	// loads csv file
	me.load = function() {
		$.ajax({
			url: 'fortune500.csv',
			context: this,
			success: function(data) {
				me.parseData(data);
				me.run();
			}
		});
	};

	me.formatDollar = function(value) {
		var add = value < 0 ? ' loss' : '';
		value = Math.abs(Number(value));
		
		return '$' + (value >= 1000 ? String(Math.round(value/100)/10).replace('\.',',')+' billion' :  String(Math.round(value*10)/10).replace('\.',',') + ' million')+add;
	};

	// parses the data structure
	me.parseData = function(dataStr) {
		var me = this;
		// each year contains a list of companies
		me.years = [];
		me.companies = {};
		var lines = dataStr.split('\n');
		var minRank = 1, maxRank = 500, 
			minRev = Number.POSITIVE_INFINITY, maxRev = Number.NEGATIVE_INFINITY,
			minRev2 = Number.POSITIVE_INFINITY, maxRev2 = Number.NEGATIVE_INFINITY,
			minProfit = Number.POSITIVE_INFINITY, maxProfit = Number.NEGATIVE_INFINITY,
			minProfit2 = Number.POSITIVE_INFINITY, maxProfit2 = Number.NEGATIVE_INFINITY;
		for (var l = 0; l < lines.length; l++) {
			if (l == 0) continue;
			var row = lines[l].split(','),
				y = row[0], rank = Number(row[1]), company = row[2],
				rev = Math.log(Number(row[3])), profit = Number(row[4]), 
				yx = y - me.minYear, defl = me.deflators[yx]*0.01,
				rev2 = Math.log(row[3]/defl), prof2 = Number(row[4]/defl);
			if (profit != 0) {
				if (Math.abs(profit) < 1) profit = 0;
				else profit = profit > 0 ? Math.log(profit) : -Math.log(-profit);
				if (Math.abs(prof2) < 1) prof2 = 0;
				else prof2 = prof2 > 0 ? Math.log(prof2) : -Math.log(-prof2);
			} 
			if (me.cutTop500 && rank > 500) continue;
			if (me.years[yx] == undefined) me.years[yx] = [];

			var compData = [rank, rev, profit, profit, rev2, prof2, prof2, company, row[4], yx, '#'+rank, me.formatDollar(row[3]), me.formatDollar(row[4])];
			

			me.years[yx].push(compData);
			
			if (me.companies[company] == undefined) {
				me.companies[company] = [];
			}
			me.companies[company].push(compData);

			if (!isNaN(rev)) {
				minRev = Math.min(minRev, rev);
				maxRev = Math.max(maxRev, rev);
				minRev2 = Math.min(minRev2, rev2);
				maxRev2 = Math.max(maxRev2, rev2);
			}
			if (!isNaN(profit)) {			
				minProfit = Math.min(minProfit, profit);
				maxProfit = Math.max(maxProfit, profit);
				minProfit2 = Math.min(minProfit2, prof2);
				maxProfit2 = Math.max(maxProfit2, prof2);
			}
		}
		me.ranges = [[minRank, maxRank], [minRev, maxRev], [minProfit, maxProfit], [0, maxProfit],
				[minRev2, maxRev2], [minProfit2, maxProfit2], [0, maxProfit2] ];
	};

	me.yByCompany = function(company) {
		var me = this, 
			c = me.curSelection,
			n = me.nextSelection,
			r = me.ranges,
			t = me.selectionProgress;
		// check for missing or cutted values
		if (company[c] == undefined 
			|| company[n] == undefined) {
			return undefined;
		}
		
		var curT = me.normalize(company, c),
			nextT = me.normalize(company, n),
			a = curT * (1-t) + nextT * t;
		return Math.round(520 - 500*a);
	};

	me.normalize = function(comp, measure) {
		var v = comp[measure], 
			me = this, 
			r = me.ranges,
			min = r[measure][0],
			max = r[measure][1],
			t = (v-min)/(max-min);
		return measure == 0 ? 1-t : t;
	};

	me.xByYear = function(year) {
		var me = this;
		return Math.round(50+(800/(me.maxYear - me.minYear)) * (year - me.minYear));
	};

	me.canvas = document.getElementById('canvas').getContext('2d');

	me.render = function() {
		var t0 = new Date();
		var me = this;
		me.curLUT = [];
		var canvas = me.canvas;
		//canvas.fillStyle = 'gray';
		canvas.clearRect(0,0,900,650);
		var bmp = canvas.getImageData(0,0,900,520);
		for (var yr = 0; yr < me.maxYear - me.minYear; yr++) {

			for (var c = 0; c < me.years[yr].length; c++) {
				var x = me.xByYear(yr+me.minYear),
					comp = me.years[yr][c],
					y = me.yByCompany(comp);
				if (y == undefined) {
					continue;
				}
				if (me.curLUT[yr] == undefined) me.curLUT[yr] = [];
				me.curLUT[yr][y] = comp;
				
				var compName = comp[7];
				
				var w = compName.length*.3;
				var alpha = 255;

				if (comp[3] < 0) {
					var cs = me.curSelection, ns = me.nextSelection;
					if ((cs == 3) && (ns == 6) || (cs == 6) && (ns == 3)) alpha = 0; 
					else if ((cs == 3 || cs == 6)) {
						alpha = Math.round(alpha*me.selectionProgress);
					} else if ((ns == 3 || ns == 6)) {
						alpha = Math.round(alpha*(1-me.selectionProgress));
					}
				}
				
				var r,g,b;
				if (me.highlightLosses && !isNaN(comp[8]) && comp[8] != 0) {
					if (comp[8] < 0) {
						r = 255; g = 140; b = 140;
					} else {
						r = b = 160; g = 225;
					}
				} else {
					r = g = b = 210;
				}

				for (var x0 = Math.round(x-w*.5);x0<x+w;x0++) {
					var idx = (x0 + y * 900) * 4;
					bmp.data[idx] = r;
					bmp.data[idx+1] = g;
					bmp.data[idx+2] = b;
					bmp.data[idx+3] = alpha;
				}
				//canvas.fillStyle = comp[me.curSelection] > 0 ? 'darkgreen' : 'darkred';
				//canvas.fillRect(Math.round(x-w*.5), y, Math.round(w), 1);
				//	from = new Point(x-w*.5,y),
				//	to = new Point(x+w*.5,y),
				//	path = new Path.Line(from, to);
				// path.strokeColor = 'gray';
								
			}
		}
		canvas.putImageData(bmp, 0, 0);
		// view.draw();
		// console.log("rendering took "+(new Date() - t0)+' ms');
	};

	me.animateTo = function(col, inflationAdjusted) {
		var me = this;
		if (col > 0 && col < 4 && inflationAdjusted === true) col+=3;
		else if (col > 3 && inflationAdjusted === false) col-=3;
		me.nextSelection = col;
		me.animateT = 0;
		//clearInterval(me.animateTimer);
		// me.animateTimer = setInterval(me.animateFrame.bind(me), 60);

		// Usage
		animLoop(me.animateFrame.bind(me), me );
	};

	me.animateFrame = function(deltaT, now) {
		var me = this;
		me.animateT += 0.06;
		me.selectionProgress = 1+Math.pow(me.animateT-1,5);
		
		if (me.selectedCompany) {
			window.f500_updatePaths(me.getPathsForCompany(me.selectedCompany));
			me.updateTooltip();
		}
		
		if (me.animateT >= 1) {
			me.selectionProgress = 1;
			me.curSelection = me.nextSelection;
			me.render();
			return false;
		}
		me.render();
		return true;
	}

	// runs the app after the data is loaded and parsed
	me.run = function() {
		//console.log("run");
		var me = this;
		me.render();
		
		// add interactivity
		//$('#papercanvas').click(me.onClick.bind(me));
		$('#papercanvas').mousemove(me.onMouseMove.bind(me));
		$('#papercanvas').mousedown(me.onMouseDown.bind(me));
		$('#papercanvas').mouseup(me.onMouseUp.bind(me));
		$('#tooltip').mouseup(me.onMouseUp.bind(me));
	};
	
	me.updateSelection = function(e) {
		var me = this, x = e.layerX, y = e.layerY;
		// get year from x
		window.f500_removePaths();
		
		
		var yr = Math.round(((x - 50)/800)*(me.maxYear - me.minYear));
		if (me.curLUT[yr][y] != undefined) {
			var comp = me.curLUT[yr][y];
			var compName = comp[7];
			var paths = me.getPathsForCompany(compName);
			window.f500_setPaths(paths);
			me.selectedCompany = compName;
			me.selectedComp = comp;
			me.selectedYear = yr;
			$('#tooltip').show();
		
			me.updateTooltip();
		} 
	};
	
	me.onClick = function(e) {
		$('#tooltip').hide();
		this.updateSelection(e);
	};
	
	me.onMouseMove = function(e) {
		var me = this;
		if (me.mouseDown === true) {
			me.updateSelection(e);
		}
	};
	
	me.onMouseDown = function(e) {
		var me = this;
		me.mouseDown = true;
		me.updateSelection(e);
	};
	
	me.onMouseUp = function(e) {
		var me = this;
		me.mouseDown = false;
	};
	
	me.updateTooltip = function() {
		var me = this, comp = me.selectedComp;
		if (!comp) return;
		$('#tooltip label.name').html(me.selectedCompany);
		var cs = me.nextSelection, y = me.yByCompany(comp);
		$('#tooltip label.value').html(comp[cs == 0 ? 10 : cs == 1 || cs == 4 ? 11 : 12]);
		$('#tooltip').css({ left: (me.xByYear(me.selectedYear+me.minYear)-10)+'px', top: 	(y+11)+'px'});
	};
	
	me.getPathsForCompany = function(compName) {
		var me = this;
		var lastYr = -1;
		var paths = [];
		var path = [];
		for (var i=0; i<me.companies[compName].length; i++) {
			var cComp = me.companies[compName][i];
			var yr = cComp[9];
			if (yr > lastYr+1) {
				// row is cut
				if (path.length > 0) paths.push(path);
				path = [];
			}
			if (yr == lastYr) continue;
			path.push([me.xByYear(yr+me.minYear), me.yByCompany(cComp)]);
			lastYr = yr;
		}
		if (path.length > 0) paths.push(path);
		return paths;
	};
	
	me.load();
};


$(function() {
	window.f500 = new Fortune500();
	$('#tooltip').hide();
});
